#!/usr/bin/perl -w
#!/opt/local/bin/perl -w

# run a taverna workflow repeatedly with a timeout and checks that it actually worked
# a bit ugly since the parameters have to be in file

# takes two arguments -- a date and a starting point

# libraries
use DBI ; 
use SOAP::Lite ; 
use XML::Simple ;
use Data::Dumper ;
use Sys::Hostname ;
use Parallel::SubFork::Task ;
use DateTime ;
use DateTime::Format::DateParse ; 
use DateTime::Format::Duration ; 
use DateTime::Format::Strptime ; 
use Net::Twitter ;
use Getopt::Long ; 

# parameters

# timezone for printing messages
use constant TZ => DateTime::TimeZone::Local->TimeZone();

# tune how batches are run
use constant INCR => 500 ;	# default for how big a chunk to run--needs to agree with workflow
use constant PAUSE => 15 ;	# pause between runs in seconds
use constant NUMRETRIES => 4 ;	# how many times to rerun a batch before going on
use constant RETRYTIME => 20 ;	# how long to wait before retrying 
use constant CHILDWAIT => 20 ; 	# how long to wait before checking child process

# where to find Taverna and the workflow
if ( hostname =~ /^floss/ ) {
  $TAVERNA = "/Applications/Taverna.app/Contents/MacOS/executeworkflow.sh" ; 
} else {
  $TAVERNA = "/Applications/\@Applications/\@Research/Taverna.app/Contents/MacOS/executeworkflow.sh" ; 
}

# signal constants
use constant SIGTERM => 15 ;    # the TERM signal to kill a subprocess

### process arguments

# defaults
$verbose = 1 ; 
$twitter = 0 ; 
$log = 1 ; 
$WORKFLOW = "summarize_v4_embedded.xml" ; # second run

# process the command line flags
GetOptions ( 'verbose|v+' => \$verbose, 
	     'twitter|t!' => \$twitter,
	     'log|l!' => \$log,
	     'workflow|w=s' => \$WORKFLOW )
  or die "Didn't understand arguments" ; 

my $usage = "Usage: run-taverna.pl date [start] [end] [inc]" ; 
# handle remaining arguments
$date = defined $ARGV[0] ? $ARGV[0] : die $usage ; 
die $usage unless $date =~ /^\d\d\d\d-\d\d/ ; 
$START = defined $ARGV[1] ? $ARGV[1] : 0 ;
die "Start must be >= 0" unless $START >= 0 ; 
$MAX   = defined $ARGV[2] ? $ARGV[2] : getprojectcount($date) ; 
die "Couldn't get end project number" unless defined $MAX ; 
$inc   = defined $ARGV[3] ? $ARGV[3] : INCR ; 


### set up run time environment

# autoflush so we can see progress
$| = 1 ; 

# avoid zombies (so they say)
#$SIG{'CHLD'} = 'IGNORE' ; 

# check required directories exist and create otherwise
use constant PARAMDIR => "params" ; 
use constant OUTPUTD => "output" ; 
use constant LOGDIR => "logs" ; 
$OUTPUTDIR = OUTPUTD . "/$date" ; 

foreach $dir ( PARAMDIR, OUTPUTD, $OUTPUTDIR, LOGDIR ) {
  if ( !-e $dir ) {
    mkdir $dir ; 
  } elsif ( !-d $dir ) { 
    die "$dir exists but is not a directory\n" ; 
  }
}

# setup taverna args
@baseargs = ($TAVERNA) ; 

# set up the date input file
open DATE, ">", PARAMDIR . "/date.txt" ; 
print DATE "$date" ; 
close DATE ; 
push @baseargs, ("-input", "date", PARAMDIR . "/date.txt") ; 

# perhaps disconnect from terminal? 

# set up log file if required
$today = DateTime->now( time_zone => TZ )->ymd ;  
if ( $log ) { 
  $logfile = LOGDIR . "/$today.log" ; 
  
  # say what we're doing if we're logging
  print "Running $WORKFLOW on projects $START to $MAX for $date logging on $logfile\n" ; 

  # redirect STDOUT 
  close STDOUT ; 
  open STDOUT, ">>", $logfile or die "Can't open logfile" ;

  # if verbose >= 2, redirect STDERR to log, otherwise dispose of it
  close STDERR ; 
  open STDERR, ">>", ($verbose >= 2 ? $logfile : "/dev/null") 
    or die "Can't open logfile" ;
}

# open a file for exception logging and an array to store exception cases
# want file so we have a record of problems in case we die and don't run END block
$exceptionslog = LOGDIR . "/exceptions-$today.log" ; 
open EXCEPTIONS, ">>", $exceptionslog ; 
@exceptions = () ; 

# at end close exceptions file and report exceptions
END { 
  close EXCEPTIONS ; 

  # try to clean up exceptions file if we didn't need it
  unlink $exceptionslog if defined $exceptionslog and -z $exceptionslog ; 

  # report exceptions if any
  if ( $#exceptions >= 0 ) { 
    print $#exceptions . " exceptions:" ; 
    print " $_" foreach @exceptions ; 
    print "\n" ; 
  }
}

# set up to twitter if requested
$TWITTER = Net::Twitter->new( netrc => 1 ) if $twitter ;

# setup a time and duration formatter
$DF = DateTime::Format::Duration->new(pattern => '%e days %H:%M:%S',
				      normalize => 1 ) ; 
$TF = DateTime::Format::Strptime->new(pattern => '%F %T') ; 

# keep track of start time to estimate finish
$STARTTIME = printlog(0, "Running $WORKFLOW on projects $START to $MAX for $date") ; 

# at end print a good bye message
END {
  my $now = DateTime->now( time_zone => TZ, formatter => $TF ) ; 
  printlog(0, ($s > $MAX ? "Finished!" : "Aborted!") . " Total run time: " .
	   $DF->format_duration($now - $STARTTIME) . "\n") ; 
}

### main loop--run taverna in chunks 

for ( $s = $START ; $s < $MAX ; $s += $inc ) { 
  # pause between runs
  if ( $s > $START ) { 
    printlog(2, "Pausing " . PAUSE . " seconds") ; 
    sleep PAUSE ;
  }

  # write the start input file
  open START,">", PARAMDIR . "/start.txt" ; 
  print START "$s" ; 
  close START ; 
  my @args = (@baseargs, "-input", "start", PARAMDIR . "/start.txt") ; 

  # file for taverna output
  my $output = "$OUTPUTDIR/$s.xml" ; 
  unlink "$output" if -e "$output" ; 
  push @args, ("-outputdoc", $output, $WORKFLOW) ;

  # more parameters based on size of chunk to run
  # NB. these may need tuning for a different workflow
  my $batchsize = $MAX - $s + 1 > $inc ? $inc : $MAX - $s + 1 ; 
  $TIMEOUT = 30 + $batchsize ;    # should finish in < 1 second per project + 30 seconds to start
  $MINOUTPUTSIZE = 500*$batchsize ; # expect to see a file of at least 500 bytes per project 

  # print a progress message and run Taverna
  printprogress($s,"Running batch") ; 
  runtaverna(@args) ; 

  my $attempts = 0 ; 
  # if running didn't create an output, sleep and try again up NUMRETRIES times
  while ( (!-e "$output" || -s "$output" < $MINOUTPUTSIZE) && $attempts++ < NUMRETRIES ) { 
    printlog(1, "Run of batch $s failed, sleeping " . RETRYTIME . " seconds") ; 
    sleep RETRYTIME ; 

    printprogress($s,"Rerun $attempts for batch") ; 
    unlink "$output" if -e "$output" ; 
    runtaverna(@args) ;
  } 

  # check if we succeeded at last, otherwise log an exception
  if ( -e "$output" && -s "$output" >= $MINOUTPUTSIZE ) {
    printlog(1, "Batch $s done") ; 
  } else {
    printlog(0, "Batch $s not done after " . NUMRETRIES . " tries") ; 
    print EXCEPTIONS "$s not done after " . NUMRETRIES . " tries\n" ; 
    push @exceptions, $s ; 
  }
}



#### SUBROUTINES

### print a log message. 
# log message has priority and message
# priority 0 == always print
#          1 == print if verbose >= 1 (progress messages)
#          2 == print if verbose >= 2 (debugging messages)
# Returns date in case we want to do more with it

sub printlog {
  my $priority = shift ; 
  my $msg = shift ; 

  my $now = DateTime->now( time_zone => TZ, formatter => $TF ) ; 

  # print log messages, depending on priority
  print "$now ($priority): $msg\n" if $priority <= $verbose ; 

  # twitter high priority messages if requested
  $TWITTER->update($msg) if $twitter && $priority == 0 ; 

  return $now ; 
}

### print a progress message 

sub printprogress {
  my $s = shift ; 
  my $status = shift ; 
  my $now = DateTime->now( time_zone => TZ ) ; 

  my $message = "$status $s/$MAX. " ; 

  my $diff = $now - $STARTTIME ; 
  $message .= "Elapsed time " . $DF->format_duration($diff) . ". " ;
  
  # if we've been going a while, print an estimate of when we'll finish  
  if ( $s > $START ) { 
    # convert days to minutes; not done automatically
    # NB will be wrong if we run over DST change, which is why it's not automatic
    my $d = DateTime::Duration->new( minutes => $diff->delta_minutes +
				                24 * 60 * $diff->delta_days, 
				     seconds => $diff->delta_seconds ) ; 

    # scale duration by proportion we've done
    # NB. trying to write this computation on one line causes type errors
    my $factor = ($MAX - $START) / ($s - $START) ; 
    $d = $d * $factor ; 
    $message .= "Estimated completion " . ($STARTTIME + $d) . ". " ; 
  }

  printlog(0, $message) ; 
}

### run taverna with a timeout and zombie reaping

sub runtaverna {
  my @args = @_ ; 

  # execute taverna in a subprocess with a timeout, passing any arguments we got
  my $task = Parallel::SubFork::Task->start(sub { exec @args ; }) ;
  printlog(2, "Started process ".$task->pid()) ; 
  my $tasktimeouttime = DateTime->now( time_zone => TZ )->add( seconds => $TIMEOUT ) ; 
  printlog(2, "Task will timeout at $tasktimeouttime") ; 

  # wait for process to finish
  # ought to be able to just wait_for($TIMEOUT) but it seems to return early
  # so check we waited long enough
  my $now = DateTime->now( time_zone => TZ ) ; 
  my $waitfor = $TIMEOUT ; 
  while ( DateTime->compare($tasktimeouttime, $now) > 0
	 && $task->wait_for($waitfor) ) {
    # if it's time, then stop
    last if DateTime->compare($tasktimeouttime, DateTime->now( time_zone => TZ )) <= 0 ; 

    # otherwise, wait a bit (less than the full duration in case task ends earlier)
    printlog(3, "Waiting $waitfor seconds 'til $tasktimeouttime\n") ;
    sleep (CHILDWAIT > $waitfor ? $waitfor : CHILDWAIT) ; 

    # and reset to wait some more
    $now = DateTime->now( time_zone => TZ ) ; 
    $diff = $tasktimeouttime - $now ;
    $waitfor = ( $diff->hours() * 60 + $diff->minutes() ) * 60 + $diff->seconds()  ; 
  } 

  # exited loop, either because task is done or because of timeout
  # check if task is still running and kill it if so
  if ( $task->wait_for(1) ) {
    printlog(2, "Execution timed out, killing process group") ; 

    # keep from killing ourselves
    local $SIG{'TERM'} = 'IGNORE' ; 

    # kill whole process group (to get shell script + java)
    kill SIGTERM, -$$ ;  
    $task->wait_for() ;	 # to be sure it died... 

    printlog(2, "Finished killing process group") ; 
  } 
}


### get number of projects over which to run given desired date

sub getprojectcount {
  # use date to determine SF schema
  my $date = shift @_ ; 
  my $d = DateTime::Format::DateParse->parse_datetime($date) 
    or die "Can't parse date $date" ;
  my $schema = "sf" . $d->strftime("%m%y") ; 

  # run the query
  my $xml = SOAP::Lite
    -> uri('http://rails-test.floss.syr.edu/notre_dame_sf/wsdl/')
      -> proxy('http://rails-test.floss.syr.edu/notre_dame_sf/api/MakeQuery')
	-> MakeQuery("username","password",
		     "$schema","count(*)","$schema.groups","is_public = 1",0)
	  -> result 
	    or die "SOAP query failed" ; 
  
  # parse the xml to get the result
  my $xdoc = XMLin($xml) or die "Could not parse SOAP query results" ;
  return $xdoc->{row}->{count} ; 
}
